--[[
						     ----o----(||)----o----(||)----o----


                                                           #########  #  #######
                                                               #      #  #      ##
                                                               #      #  #        #
                                                               #      #  #      ##
                                                               #      #  #######
                                                               #      #  #
                                                               #      #  #
                                                               #      #  #
                                                               #      #  #


						    ----o----(||)----oo----(||)----o----

							  v2.02 3rd September 2017
						    Copyright (C) Taraezor / Chris Birch

						    ----o----(||)----oo----(||)----o----

]]
--------------------------------------------------------------------------------------------------------------------------------------------

local addonName, L = ...

local gmatch = string.gmatch
local lower = string.lower
local match = string.match
local print = print
local random = random
local select = select
local sub = string.sub
local strsplit = strsplit
local tonumber = tonumber
local tostring = tostring

local GetActionInfo = GetActionInfo
local GetActiveSpecGroup = GetActiveSpecGroup
local GetAuctionItemBattlePetInfo = GetAuctionItemBattlePetInfo
local GetAuctionItemInfo = GetAuctionItemInfo
local GetAuctionItemLink = GetAuctionItemLink
local GetBuybackItemLink = GetBuybackItemLink
local GetCurrencyListLink = GetCurrencyListLink
local GetFactionInfo = GetFactionInfo
local GetFriendshipReputation = GetFriendshipReputation
local GetItemSpell = GetItemSpell
local GetMerchantItemLink = GetMerchantItemLink
local GetPetInfoByPetID = C_PetJournal.GetPetInfoByPetID
local GetPetInfoBySpeciesID = C_PetJournal.GetPetInfoBySpeciesID
local GetPvpTalentInfoByID = GetPvpTalentInfoByID
local GetQuestLogItemLink = GetQuestLogItemLink
local GetTalentInfoByID = GetTalentInfoByID
local PlaySound = PlaySound
local StopSound = StopSound
local TransmogCollectionGetSourceInfo = C_TransmogCollection.GetSourceInfo
local TransmogSetsGetSourcesForSlot = C_TransmogSets.GetSourcesForSlot
local UnitAura = UnitAura
local UnitBuff = UnitBuff
local UnitDebuff = UnitDebuff
local UnitFactionGroup = UnitFactionGroup
local UnitGUID = UnitGUID

--------------------------------------------------------------------------------------------------------------------------------------------
-- Print Colours - Tip Theme
--------------------------------------------------------------------------------------------------------------------------------------------

local pc_colour_Prefix		= "\124cFF0070DE"
local pc_colour_Highlight	= "\124cFFB19EFF"
local pc_colour_PlainText	= "\124cFF00B7EB"
local prefix = "\124cFF0070DETip\124cFF00B7EB "

local function printPC( message )
	DEFAULT_CHAT_FRAME:AddMessage( pc_colour_Prefix.. addonName.. ": ".. pc_colour_PlainText.. message.. "\124r" )
end

--==========================================================================================================================================
--
--							Battle Pet 
--
--==========================================================================================================================================

--------------------------------------------------------------------------------------------------------------------------------------------
-- Battle Pet Auctions - Replace Blizzard's Global Function
--------------------------------------------------------------------------------------------------------------------------------------------

local BATTLE_PET_TOOLTIP = {}
local creatureIndexID = 0		-- Set in Tip_TooltipSetAuctionItem

function BattlePetToolTip_Show( speciesID, level, breedQuality, maxHealth, power, speed, customName )

	if (speciesID and speciesID > 0) then
		local name, icon, petType = C_PetJournal.GetPetInfoBySpeciesID(speciesID);
		BATTLE_PET_TOOLTIP.speciesID = speciesID;
		BATTLE_PET_TOOLTIP.name = name;
		BATTLE_PET_TOOLTIP.level = level;
		BATTLE_PET_TOOLTIP.breedQuality = breedQuality;
		BATTLE_PET_TOOLTIP.petType = petType;
		BATTLE_PET_TOOLTIP.maxHealth = maxHealth;
		BATTLE_PET_TOOLTIP.power = power;
		BATTLE_PET_TOOLTIP.speed = speed;
		if (customName ~= BATTLE_PET_TOOLTIP.name) then
			BATTLE_PET_TOOLTIP.customName = customName;
		else
			BATTLE_PET_TOOLTIP.customName = nil;
		end

		BattlePetTooltipTemplate_SetBattlePet(BattlePetTooltip, BATTLE_PET_TOOLTIP);

		local owned = C_PetJournal.GetOwnedBattlePetString(speciesID);

		-- My hack is here
		-- I noted the hard coded heights so just added 14. Normally we should get the text height and add 2 for line spacing
		local creatureID = GetAuctionItemBattlePetInfo( "list", creatureIndexID )
		if( owned == nil ) then
			BattlePetTooltip.Owned:SetText( pc_colour_Prefix.. addonName.. " \124cFF00B7EBNPC ID: \124cFF93CCEA".. 
				( creatureID or "-" ).. "\124cFF00B7EB   Species ID: \124cFF93CCEA".. ( speciesID or "-" ).. "\124r" )
			BattlePetTooltip:SetSize( 260, 136 )
		else
			BattlePetTooltip.Owned:SetWordWrap( true )
			BattlePetTooltip.Owned:SetText( owned.. "\n".. pc_colour_Prefix.. addonName.. " \124cFF00B7EBNPC ID: \124cFF93CCEA".. 
				( creatureID or "-" ).. "\124cFF00B7EB   Species ID: \124cFF93CCEA".. ( speciesID or "-" ).. "\124r" )
			BattlePetTooltip:SetSize( 260, 150 )
		end

		BattlePetTooltip:Show();

		BattlePetTooltip:ClearAllPoints();
		BattlePetTooltip:SetPoint(GameTooltip:GetPoint());
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Chat linked Reference - Companion and Battle Pets (I also show the IDs in chat)
--------------------------------------------------------------------------------------------------------------------------------------------

-- Note that I haven't tried the same as with BattlePetToolTip_Show. More extensive hack so liable to breaking?

local savedBattlePetID, savedSpeciesID;

function Tip_FloatingBattlePetShow( self, _, _, _, _, _, _, battlePetID )

	if ( savedBattlePetID ~= FloatingBattlePetTooltip.battlePetID ) and ( savedSpeciesID ~= FloatingBattlePetTooltip.speciesID ) then

		if ( tostring( FloatingBattlePetTooltip.battlePetID ) == "0000000000000000" ) then
			local _, _, _, npcID, _, _, _, _, _, _, _, displayID = GetPetInfoBySpeciesID( FloatingBattlePetTooltip.speciesID )
			printPC( "Species ID: \124cFF93CCEA".. FloatingBattlePetTooltip.speciesID.. "\n"..
				"\124cFF00B7EBDisplay ID: \124cFF93CCEA".. displayID..
				"\124cFF00B7EB   NPC ID: \124cFF93CCEA".. tostring( npcID or "nil" ) )
		else
			local speciesID, _, _, _, _, displayID, _, _, _, _, npcID = GetPetInfoByPetID( FloatingBattlePetTooltip.battlePetID )
			printPC( "Species ID: \124cFF93CCEA".. speciesID.. "\n"..
				"\124cFF00B7EBDisplay ID: \124cFF93CCEA".. displayID..
				"\124cFF00B7EB   NPC ID: \124cFF93CCEA".. tostring( npcID or "nil" ).. "\n"..
				"\124cFF00B7EBPet ID: \124cFF93CCEA".. FloatingBattlePetTooltip.battlePetID.. "\n" )
		end
		savedBattlePetID = FloatingBattlePetTooltip.battlePetID
		savedSpeciesID = FloatingBattlePetTooltip.speciesID
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Pet Journal Pet Card
--------------------------------------------------------------------------------------------------------------------------------------------

local function Tip_PetJournalUpdatePetCard( self )

	if ( PetJournalPetCard.petID == nil ) then
		local _, _, _, npcID, _, _, _, _, _, _, _, displayID = GetPetInfoBySpeciesID( PetJournalPetCard.speciesID )
		if ( TipDB.blankPrefixLine == "y" ) then
			GameTooltip:AddLine( "\n".. prefix.. "Species ID: \124cFF93CCEA".. PetJournalPetCard.speciesID.. "\n"..
					"\124cFF00B7EBDisplay ID: \124cFF93CCEA".. displayID..
					"\124cFF00B7EB   NPC ID: \124cFF93CCEA".. npcID )

		else
			GameTooltip:AddLine( prefix.. "Species ID: \124cFF93CCEA".. PetJournalPetCard.speciesID.. "\n"..
					"\124cFF00B7EBDisplay ID: \124cFF93CCEA".. displayID..
					"\124cFF00B7EB   NPC ID: \124cFF93CCEA".. npcID )
		end
	else
		local speciesID, _, _, _, _, displayID, _, _, _, _, npcID = GetPetInfoByPetID( PetJournalPetCard.petID )
		if ( TipDB.blankPrefixLine == "y" ) then
			GameTooltip:AddLine( "\n".. prefix.. "Species ID: \124cFF93CCEA".. speciesID.. "\n"..
					"\124cFF00B7EBDisplay ID: \124cFF93CCEA".. displayID..
					"\124cFF00B7EB   NPC ID: \124cFF93CCEA".. npcID.. "\n"..
					"\124cFF00B7EBPet ID: \124cFF93CCEA".. PetJournalPetCard.petID.. "\n" )
		else
			GameTooltip:AddLine( prefix.. "Species ID: \124cFF93CCEA".. speciesID.. "\n"..
					"\124cFF00B7EBDisplay ID: \124cFF93CCEA".. displayID..
					"\124cFF00B7EB   NPC ID: \124cFF93CCEA".. npcID.. "\n"..
					"\124cFF00B7EBPet ID: \124cFF93CCEA".. PetJournalPetCard.petID.. "\n" )
		end
	end
	GameTooltip:Show()
end

--==========================================================================================================================================
--
--							Add Tooltip Line
--
--==========================================================================================================================================

local function AddLine( type, entityID, passedTooltip )

	if ( entityID == nil ) or ( entityID == 0 ) then return end

	local tooltip = passedTooltip or GameTooltip
	local tooltipName = tooltip:GetName()

	local line, found = "", false
	local compareLine = prefix.. type.. " ID:"
	for i = tooltip:NumLines(), 2, -1 do
		line = _G[ tooltipName.. "TextLeft"..i ]:GetText()
		if ( line ~= nil ) then
			if ( match( line, compareLine ) ) then
				found = true
				break
			end
		end
	end
	if ( found == false ) then
		compareLine = ( TipDB.blankPrefixLine == "y" ) and ( "\n".. compareLine ) or compareLine
		tooltip:AddLine( compareLine.. " \124cFF93CCEA".. entityID )
		tooltip:Show()
		return true
	else
		return false
	end
end

local function GetCurrencyFromTooltip()

	local currency = GameTooltipTextLeft1:GetText()
	for i = 1, GetCurrencyListSize() do
		local link = GetCurrencyListLink( i ) 
		if link then
			local curID, curName = match( link, "|?H?[^:]*:?(%d+)|?h?%[?([^%[%]]*)%]?|?h?|?r?" )
			if ( curName == currency ) then
				return AddLine( "Currency", curID )
			end
		end
	end

	local itemLink = select( 2, GameTooltip:GetItem() )
	if ( itemLink ~= nil ) then
		local spellItemID = tonumber( match( itemLink, "item:(%d+):" ) )
		if ( spellItemID ~= nil ) then
			AddLine( "Item", spellItemID )
			return true
		end
	end

	return false
end

local function GetItemFromLink( link, passedTooltip )

	local tooltip = passedTooltip or GameTooltip

	local itemLink = link or select( 2, tooltip:GetItem() )
	if ( itemLink ~= nil ) then

		local spellItemID = tonumber( match( itemLink, "item:(%d+):" ) )
		if ( spellItemID ~= nil ) then
			AddLine( "Item", spellItemID, tooltip )
			return true
		end
	end

	-- Wasn't an item so should probably be a currency

	if ( tooltip:GetName() ~= "GameTooltip" ) then return false end

	local currency = GameTooltipTextLeft1:GetText()
	if not currency then return false end

	for i = 1, GetCurrencyListSize() do
		local link = GetCurrencyListLink( i ) 
		if link then
			local curID, curName = match( link, "|?H?[^:]*:?(%d+)|?h?%[?([^%[%]]*)%]?|?h?|?r?" )
			if ( curName == currency ) then
				return AddLine( "Currency", curID )
			end
		end
	end
end

local function GetSpellFromTooltip()
	local spellItemID = select( 3, GameTooltip:GetSpell() )
	if ( spellItemID ~= nil ) then
		AddLine( "Spell", spellItemID )
		return true
	end
	return false
end

--==========================================================================================================================================
--
--							Hyperlinks
--
--==========================================================================================================================================

--------------------------------------------------------------------------------------------------------------------------------------------
-- Anything supported by GameTooltipSetHyperlink
--	* Artifact Relics
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_GameTooltipSetHyperlink( self, link )

	local idType, spellItemID = match( link, "(%w+):(%d+)" )
	local label = sub( idType, 2, 2):upper().. sub( idType, 3 )
	AddLine( label, spellItemID )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Chat linked References for anything supported by ItemRefTooltipSetHyperlink
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_ItemRefTooltipSetHyperlink( self, link )

	local idType, spellItemID = match( link, "(%w+):(%d+)" )
	local label = sub( idType, 1, 1):upper().. sub( idType, 2 )
	AddLine( label, spellItemID, ItemRefTooltip )

	if idType == "talent" then
		local target = UnitName("target") and true or false
		local activeSpec = GetActiveSpecGroup( target )
		local talentSpellID = select( 6, GetTalentInfoByID( spellItemID, activeSpec ) )
		AddLine( "Spell", talentSpellID, ItemRefTooltip )
	elseif idType == "pvptal" then
		local pvpTalentSpellID = select( 6, GetPvpTalentInfoByID( spellItemID ) )
		AddLine( "Spell", pvpTalentSpellID, ItemRefTooltip )
	elseif idType == "apower" then
		local artifactInfo = C_ArtifactUI.GetPowerInfo( spellItemID )
		AddLine( "Spell", artifactInfo.spellID, ItemRefTooltip )
	end
	ItemRefTooltip:Show()
end

--==========================================================================================================================================
--
--							Event Traps
--
--==========================================================================================================================================

--------------------------------------------------------------------------------------------------------------------------------------------
-- NPC Friendship Status Bars (E.g. The Tillers members)
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_NPCFriendshipStatusBar()
	local factionID = select( 1, GetFriendshipReputation() )
	AddLine( "Faction / Friendship", factionID )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Reputation Tab of the Character Frame
--------------------------------------------------------------------------------------------------------------------------------------------

local function Tip_ReputationBar( self )

	if self.friendshipID then
		AddLine( "Faction / Friendship", self.friendshipID )
	else
		if not GameTooltipTextLeft1:GetText() then
			GameTooltip:SetOwner( self, "ANCHOR_BOTTOMRIGHT" )
		end
		local factionID = select( 14, GetFactionInfo( self.index ) )
		AddLine( "Faction", factionID )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Vendor Gear Items - mousing over causes "shopping" tooltips to popup
--------------------------------------------------------------------------------------------------------------------------------------------

local function Tip_ShoppingTooltip_OnTooltipSetItem( self )
	GetItemFromLink( nil, self )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- World Map Tooltips - Note that this does the numbered button but NOT the blob
-- Legion introduced "Available Quests" along with two new API functions. Sadly, I cannot extract a quest ID. They appear as "!" on the map
--------------------------------------------------------------------------------------------------------------------------------------------

local function Tip_WorldMapTooltip( self, arg1 )

	local idType, idValue = "Quest ID:";
	if self:GetOwner().questID then
		idValue = self:GetOwner().questID
	else
		idType = "Quest Line:"
	end

	local line, found = "", false
	local compareLine = prefix.. idType
	for i = WorldMapTooltip:NumLines(), 2, -1 do
		line = _G["WorldMapTooltipTextLeft"..i ]:GetText()
		if ( line ~= nil ) then
			if ( match( line, compareLine ) ) then
				found = true
				break
			end
		end
	end
	if ( found == true ) then
		return
	end

	if not self:GetOwner().questID then
		local questTitle = WorldMapTooltipTextLeft1:GetText()
		for i = 1, C_Questline.GetNumAvailableQuestlines() do
			local questLine, questName = C_Questline.GetQuestlineInfoByIndex( i )
			if ( questTitle == questName ) then
				idValue = questLine
				found = true
				break
			end
		end
		if ( found == false ) then
			return
		end
	end

	compareLine = ( TipDB.blankPrefixLine == "y" ) and ( "\n".. compareLine ) or compareLine
	WorldMapTooltip:AddLine( compareLine.. " \124cFF93CCEA".. idValue )
	WorldMapTooltip:Show()
end

--==========================================================================================================================================
--
--							Secure Script Hooks
--
--==========================================================================================================================================

--------------------------------------------------------------------------------------------------------------------------------------------
-- Action Bar Button
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetAction( self, buttonNum )

	local actionType, spellItemID, subType = GetActionInfo( buttonNum )
	if actionType == nil then return end
	AddLine( sub( actionType, 1, 1):upper().. sub( actionType, 2 ), spellItemID )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Auctions - Items for sale, whether by the player or someone else
--------------------------------------------------------------------------------------------------------------------------------------------

-- Can't just call GetItemFromLink() as the GameTooltip doesn't have an itemLink stored for battle pets or recipes

function Tip_TooltipSetAuctionItem( self, type, index )

	local auctionType = ""
	local frameName = GetMouseFocus():GetName()
	if sub( frameName, 1, 14 ) == "AuctionsButton" then
		auctionType = "owner"
	elseif sub( frameName, 1, 12 ) == "BrowseButton" then
		auctionType = "list"
	else
		auctionType = "bidder"
	end

	local link = GetAuctionItemLink( auctionType, index )
	if GetItemFromLink( link ) then
		-- Only Recipes require the GetAuctionItemLink to be passed. Battle Pets will always fail regardless
		return
	end

	local itemID = select( 17, GetAuctionItemInfo( auctionType, index ) )
	if itemID == 82800 then			-- Pet Cage!
		creatureIndexID = index		-- Used in BattlePetToolTip_Show (above)
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Auctions - Items being setup for sale and not yet posted for auction
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetAuctionSellItem( self )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Bank and Backpack Bags
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetBagItem( self, container, slot )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Merchant Buyback Item
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetBuybackItem( self, itemIndex )
	local link = GetBuybackItemLink( itemIndex )
	GetItemFromLink( link )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Currency - Frames where a currency cost is shown
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetCurrencyByID( self, currencyID )
	AddLine( "Currency", currencyID )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Currency - The Currency Tab of the Character Frame
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetCurrencyToken( self, index )
	local link = match( GetCurrencyListLink( index ), "currency:(%d+)" )
	AddLine( "Currency", link )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Gem Socket (already set)
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetExistingSocketGem( self, index )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Guild Bank
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetGuildBankItem( self, parm1, parm2 )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Heirloom tab of the Collections frame
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetHeirloomByItemID( self, spellItemID )
	if ( spellItemID ~= nil ) then
		AddLine( "Spell", spellItemID )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Inbox & Open Mail Attachment items
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetInboxItem( self, mailIndex, attachmentIndex )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Inventory Item of player/target Character Frame; Main Bank and Reagent Bank Frames
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetInventoryItem( self, unitID, slot )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Game Time Tokens
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetItemByID( self, itemID  )
	AddLine( "Item", itemID )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Loot Frame Items
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetLootItem( self, slot )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Merchant Trade Item
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetMerchantItem( self, itemIndex )
	local link = GetMerchantItemLink( itemIndex )
	GetItemFromLink( link )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Mounts pane of the Collections Frame
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetMountBySpellID( self, spellItemID )
	AddLine( "Spell", spellItemID )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- PvP Talent Pane of the Talent Frame
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetPvpTalent( self, pvpTalentID )
	local link = select( 6, GetPvpTalentInfoByID( pvpTalentID ) )
	AddLine( "Spell", link )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Quest Currency Rewards (prior to accepting the quest)
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetQuestCurrency( self, type, itemIndex )

	local link = GetQuestItemLink( type, itemIndex )
	GetItemFromLink( link )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Quest Items Rewards (prior to accepting the quest) & Required Items (prior to handing in)
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetQuestItem( self, type, itemIndex )

	local link = GetQuestItemLink( type, itemIndex )
	GetItemFromLink( link )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Quest Log Currency
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetQuestLogCurrency( self, type, itemIndex )
	GetCurrencyFromTooltip()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Quest Log Items Rewards (an accepted quest with fixed and/or choice of rewards)
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetQuestLogItem( self, type, itemIndex )

	local link = GetQuestLogItemLink( type, itemIndex )
	GetItemFromLink( link )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Special Use Quest Items (clickable and displayed on the right side of the UI)
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetQuestLogSpecialItem( self )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Quest Reward - DEBUG - Can't find an available example to test this lol
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetQuestRewardSpell( self, spellIndex )

	local spellItemID = select( 3, self:GetSpell() )
	AddLine( "Spell", spellItemID )
print("here")
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Profession Skill List Pane - Results & Reagents
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetRecipeItem( self, recipeID, reagentIndex )

	local link = ( reagentIndex == nil ) and C_TradeSkillUI.GetRecipeItemLink( recipeID ) or
				C_TradeSkillUI.GetRecipeReagentItemLink( recipeID, reagentIndex )
	if ( link ~= nil ) then
		local idType, spellItemID = match( link, "|H(%w+):(%d+)" )
		if ( spellItemID ~= nil ) then
			if ( idType == "item" ) then
				AddLine( "Item", spellItemID )
			else
				AddLine( "Spell", spellItemID )	-- e.g. Leather Working's Primal Leg Reinforcements is "enchant"
			end
		end
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Send Mail Attachment items
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetSendMailItem( self, attachmentIndex )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Player's Stance Button
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetShapeshift( self, index )
	GetSpellFromTooltip()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Gem Socket (placed into socket but not yet "set")
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetSocketGem( self, index )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Spellbook & Professions Panes and Spellbook
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetSpellBookItem( self, spellbookIndex, bookType )
	GetSpellFromTooltip()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Spell Flyouts - E.g. Mage Portals, whether in Spell Book or on an Action Bar
-- Quest Rewards for offered quests
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetSpellByID( self, spellItemID )
	AddLine( "Spell", spellItemID )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Talent Pane of the player/target Talent Frame
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetTalent( self, ... )
	GetSpellFromTooltip()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Toybox of the Collections Frame
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetToyByItemID( self, spellItemID )
	AddLine( "Item", spellItemID )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Trainer Pane
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetTrainerService( self, index )
	GetSpellFromTooltip()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- OnUpdate Handler
--------------------------------------------------------------------------------------------------------------------------------------------

-- Typically this is where I put hacks until I can find an appropriate function to hook. Sigh. Careful as may cause duplicates

local prevUpdate = 0
 
function Tip_OnUpdate( self, elapsed )

	-- Note that Blizzard themselves use '( prevUpdate < ( self.updateTooltip or TOOLTIP_UPDATE_TIME ) )' but this causes flicker for me

	prevUpdate = prevUpdate + elapsed
	if ( prevUpdate < TipDB.updateInterval ) then
		return
	end
	prevUpdate = 0

	local owner = GameTooltip:GetOwner()
	if ( owner == nil ) then return end
	local ownerName = owner:GetName()

	if ( ownerName == "UIParent" ) then
		local _, unitID = GameTooltip:GetUnit()
		if ( unitID ~= nil ) then
			local guID = UnitGUID( unitID )
			if ( guID ~= nil ) then
				local type, _, _, _, _, npcID, _ = strsplit( "-", guID )
				if ( type == "Player" ) or ( type == "Vignette" ) or ( npcID == nil) then
				else
					local line = _G["GameTooltipTextLeft"..GameTooltip:NumLines() ]:GetText()
					if ( line ~= prefix.. "NPC ID: \124cFF93CCEA".. npcID ) then
						AddLine( "NPC", npcID )
					end
				end
			end
		end

	elseif ( ownerName == addName ) then

		-- Specialization Tab of the Talent Frame & The Artifact Frame

		if owner.spellID then
			AddLine( "Spell", owner.spellID )
			return
		end

		local parent = owner:GetParent()

		if WardrobeCollectionFrame and WardrobeCollectionFrame:IsVisible() then

			local sourceID, sourceInfo;

			-- Appearances (and Illusions) Wardrobe Collections

			if ( WardrobeCollectionFrame.selectedTab == 1 ) then
				if ( parent.transmogType == LE_TRANSMOG_TYPE_APPEARANCE ) then
					-- Allow for player selecting "hidden", e.g. for the head slot
					if WardrobeCollectionFrame.tooltipSourceIndex == nil then return end
					-- :GetActiveCategory() existence is another way to avoid enchants/illusions
					-- Can't test for active slot as it shows up as mainhandslot / secondaryhandslot - same as weapon
					local sources = WardrobeCollectionFrame_GetSortedAppearanceSources( parent.tooltipVisualID )
					sourceID = sources[ WardrobeCollectionFrame.tooltipSourceIndex ].sourceID
					sourceInfo = TransmogCollectionGetSourceInfo( sourceID )
					local spellItemID = sourceInfo.itemID
					AddLine( "Visual", ( owner.visualInfo.visualID or "-").. 
						"\124cFF00B7EB   Source ID: \124cFF93CCEA".. ( sourceID or "-").. "\n".. 
						"\124cFF00B7EB   Item ID: \124cFF93CCEA".. ( spellItemID or "-").. 
						"\124cFF00B7EB   Source Index: \124cFF93CCEA".. ( WardrobeCollectionFrame.tooltipSourceIndex or "-").. "\n"..
						"\124cFF00B7EB   Source Type: \124cFF93CCEA".. ( sourceInfo.sourceType or "-")..
						"\124cFF00B7EB   Inv Type: \124cFF93CCEA".. ( sourceInfo.invType or "-").. "\n"..
						"\124cFF00B7EB   Item Mod ID: \124cFF93CCEA".. ( sourceInfo.itemModID or "-")..
						"\124cFF00B7EB   Category ID: \124cFF93CCEA".. ( sourceInfo.categoryID or "-") )
				elseif ( parent.transmogType == LE_TRANSMOG_TYPE_ILLUSION ) then
					local visualsList = parent:GetFilteredVisualsList()
					for i = 1, #visualsList do
						if ( visualsList[i].visualID ==  owner.visualInfo.visualID ) then
							sourceID = visualsList[i].sourceID
							break
						end
					end
					sourceInfo = TransmogCollectionGetSourceInfo( sourceID )
					AddLine( "Visual", ( owner.visualInfo.visualID or "-").. 
						"\124cFF00B7EB   Source ID: \124cFF93CCEA".. ( sourceID or "-").. "\n".. 
						"\124cFF00B7EB   Source Type: \124cFF93CCEA".. ( sourceInfo.sourceType or "-")..
						"\124cFF00B7EB   Inv Type: \124cFF93CCEA".. ( sourceInfo.invType or "-").. "\n"..
						"\124cFF00B7EB   Item Mod ID: \124cFF93CCEA".. ( sourceInfo.itemModID or "-")..
						"\124cFF00B7EB   Category ID: \124cFF93CCEA".. ( sourceInfo.categoryID or "-") )
				end
				return

			-- Wardrobe Collection Appearances Sets

			elseif ( WardrobeCollectionFrame.selectedTab == 2 ) then

				local parentParent = parent:GetParent()
				local sources = TransmogSetsGetSourcesForSlot( parentParent.selectedSetID, parentParent.tooltipTransmogSlot )
				local sourceInfo = sources[parentParent:GetParent().tooltipSourceIndex]
				AddLine( "Visual", ( sourceInfo.visualID or "-").. 
					"\124cFF00B7EB   Source ID: \124cFF93CCEA".. ( sourceInfo.sourceID or "-").. "\n"..
					"\124cFF00B7EB   Item ID: \124cFF93CCEA".. ( sourceInfo.itemID or "-").. 
					"\124cFF00B7EB   Source Index: \124cFF93CCEA".. ( parentParent:GetParent().tooltipSourceIndex or "-").. "\n"..
					"\124cFF00B7EB   Source Type: \124cFF93CCEA".. ( sourceInfo.sourceType or "-")..
					"\124cFF00B7EB   Inv Type: \124cFF93CCEA".. ( sourceInfo.invType or "-").. "\n"..
					"\124cFF00B7EB   Item Mod ID: \124cFF93CCEA".. ( sourceInfo.itemModID or "-")..
					"\124cFF00B7EB   Category ID: \124cFF93CCEA".. ( sourceInfo.categoryID or "-").. "\n"..
					"\124cFF00B7EB   Set ID: \124cFF93CCEA".. ( parentParent.selectedSetID or "-") )
			end

		elseif owner.questID then
			AddLine( "Quest", owner.questID )

		else
			local parentName = parent and parent:GetName() or "nil"
			if ( parentName == "BankSlotsFrame" ) or ( parentName == "CharacterStatsPane" ) or
				( sub( parentName, 1, 11 ) == "TokenFrameC" ) then
				return
			elseif ( TipDB.debug == "y" ) then
				print( "ownerName == addName left overs, parentName:".. parentName )
			end
		end

	elseif ( self:GetName() == "ShoppingTooltip1" ) or ( self:GetName() == "ShoppingTooltip2" ) then
		return

	elseif owner.objectType then

		if ( owner.objectType == "spell" ) then
			AddLine( "Spell", owner.spellID )
			if ( TipDB.debug == "y" ) then print("new spell handler") end
		elseif ( owner.objectType == "currency" ) or ( owner.objectType == "item" ) then
			return
		else
			if ( TipDB.debug == "y" ) then print("OnUpdate owner.objectType=".. owner.objectType) end
		end

	elseif match( ownerName, "AltCurrency" ) then		-- Vendor special currencies
		GetCurrencyFromTooltip()

	elseif ( sub( ownerName, 1, 10 ) == "LootButton" ) then
		if not GetItemFromLink() then GetCurrencyFromTooltip() end

	elseif ( sub( ownerName, 1, 29 ) == "ClassTrainerScrollFrameButton" ) then
		AddLine( "Spell", owner.spellID )

	elseif ( sub( ownerName, 1, 65 ) == "PlayerTalentFrameSpecializationSpellScrollFrameScrollChildAbility" ) then
		-- The six or so spells, with alternate tooltip text, appearing in the Specialization tab of the Talents Frame
		AddLine( "Spell", owner.spellID )

	else
		-- CHAT SPAM - Note some things already processed will drop through to here so careful

		if ( TipDB.debug == "y" ) then 
			if ( sub( ownerName, 1, 24 ) == "PlayerTalentFrameTalents" ) or ( sub( ownerName, 1, 9 ) == "Character" ) or
				( sub( ownerName, 1, 11 ) == "SpellButton" ) or ( sub( ownerName, 1, 20 ) == "SpellBookSkillLineTa" ) or
				( sub( ownerName, 1, 20 ) == "MountJournalSummonRa" ) or ( sub( ownerName, 1, 8 ) == "MultiBar" ) or 
				( sub( ownerName, 1, 12 ) == "ActionButton" ) or ( sub( ownerName, 1, 13 ) == "ReputationBar" ) or
				( sub( ownerName, 1, 20 ) == "AchievementFrameCateg" ) or ( sub( ownerName, 1, 8 ) == "SendMail" ) or
				( sub( ownerName, 1, 14 ) == "ContainerFrame" ) or ( sub( ownerName, 1, 11 ) == "ReagentBank" ) or
				( sub( ownerName, 1, 9 ) == "BankFrame" ) or ( sub( ownerName, 1, 8 ) == "BankItem" ) or
				( sub( ownerName, 1, 10 ) == "BuffButton" ) or ( sub( ownerName, 1, 11 ) == "TempEnchant" ) or
				( sub( ownerName, 1, 11 ) == "PrimaryProf" ) or ( sub( ownerName, 1, 13 ) == "SecondaryProf" ) or
				( sub( ownerName, 1, 8 ) == "Merchant" ) or ( sub( ownerName, 1, 9 ) == "PaperDoll" ) or
				( sub( ownerName, 1, 10 ) == "TaxiButton" ) or ( sub( ownerName, 1, 8 ) == "MailItem" ) or
				( sub( ownerName, 1, 11 ) == "CalendarDay" ) or 
				( sub( ownerName, 1, 11 ) == "PlayerFrame" ) or ( sub( ownerName, 1, 12 ) == "ClassTrainer" ) then
				return
			else
				print( "Catch all ownerName=".. (ownerName or "nil") )
			end
		end
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Targeted Units - only when mouse over a portrait
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetUnit( self, unitID )

	local guID = UnitGUID( unitID )
	if ( guID ~= nil ) then
		local type, _, _, _, _, npcID, _ = strsplit( "-", guID )
		if ( type == "Player" ) or ( type == "Vignette" ) or ( npcID == nil ) then
		else
			AddLine( "NPC", npcID )
		end
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Player's Buffs and Debuffs
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetUnitAura( self, unitID, auraIndex, filter )

	local spellItemID = select( 11, UnitAura( unitID, auraIndex, filter ) )
	if ( spellItemID ~= nil ) then
		AddLine( "Spell", spellItemID )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Target's Buffs
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetUnitBuff( self, unitID, auraIndex )

	local spellItemID = select( 11, UnitBuff( unitID, auraIndex ) )
	if ( spellItemID ~= nil ) then
		AddLine( "Spell", spellItemID )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Target's Debuffs
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetUnitDebuff( self, unitID, auraIndex )

	local spellItemID = select( 11, UnitDebuff( unitID, auraIndex ) )
	if ( spellItemID ~= nil ) then
		AddLine( "Spell", spellItemID )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Item Upgrade
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetUpgradeItem( self )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Void Storage item to be deposited
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetVoidDepositItem( self, index )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Void Storage item already deposited
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetVoidItem( self, one, locationIndex ) -- "one" because it always had a value of "1" in testing
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Void Storage item to be withdrawn
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetVoidWithdrawalItem( self, index )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Hooks for after most/all of the UI has been loaded
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_OnEvent( self, event, parm1 )

	if ( event == "ADDON_LOADED") and ( parm1 == "Blizzard_Collections" ) then
		PetJournalPetCardPetInfo:HookScript( "OnEnter", Tip_PetJournalUpdatePetCard )

	elseif ( event == "ADDON_LOADED") and ( parm1 == addonName ) then
		local addLine = tostring( TipDB.blankPrefixLine )
		local updateInterval = TipDB.updateInterval or 0.05
		local debug = TipDB.debug or "n"
		addLine = ( addLine == "true" or addLine == "y" ) and "y" or "n"
		TipDB.blankPrefixLine = addLine
		TipDB.updateInterval = updateInterval
		TipDB.debug = debug

		NPCFriendshipStatusBar:HookScript( "OnEnter", Tip_NPCFriendshipStatusBar )

		-- No other way to do this. Each must be done individually
		for i = 1, 15 do
			_G["ReputationBar"..i]:HookScript( "OnEnter", Tip_ReputationBar )
		end

		WorldMapTooltip:HookScript( "OnUpdate", Tip_WorldMapTooltip )

		ShoppingTooltip1:HookScript( "OnTooltipSetItem", Tip_ShoppingTooltip_OnTooltipSetItem )
		ShoppingTooltip2:HookScript( "OnTooltipSetItem", Tip_ShoppingTooltip_OnTooltipSetItem )
	end

end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Slash Command / Configuration Options
--------------------------------------------------------------------------------------------------------------------------------------------

local soundkitHandle = nil

SLASH_Tip1 = "/tip"

SlashCmdList[ "Tip" ] = function( options )

	local secondParm, options1
	for v in gmatch( options, "(%S+)" ) do
		v = lower(v)
		if not options1 then
			options1 = sub( v,1,1 )
		else
			secondParm = sub( v,1,5 )
			break
		end
	end

	if ( options1 == "?" ) or not options1 then
		printPC( pc_colour_Highlight.. L["Commands"].. ":\n"..
			pc_colour_Highlight.. "a".. pc_colour_PlainText.. " = ".. L["Add an empty line"].. "? (".. L["Toggle"].. ")\n"..
			pc_colour_Highlight.. "u".. pc_colour_PlainText.. " = ".. L["Update Interval"].. " ( >= 0, <= 0.5 )\n"..
			pc_colour_Highlight.. "v".. pc_colour_PlainText.. " = ".. L["Show the version number"].. "\n")

	elseif ( options1 == "a" ) then
		if secondParm then
			printPC( "/tip a (".. ( ( TipDB.blankPrefixLine == "y" ) and L["On"] or L["Off"] ).. ")" )
		else
			TipDB.blankPrefixLine = ( TipDB.blankPrefixLine == "y" ) and "n" or "y"
			if ( TipDB.blankPrefixLine == "y" ) then
				printPC( L["Add an empty line"].. ": ".. pc_colour_Highlight.. L["On"] )
			else
				printPC( L["Add an empty line"].. ": ".. pc_colour_Highlight.. L["Off"] )
			end
		end

	elseif ( options1 == "u" ) then
		local updateInterval = tonumber( secondParm or 100 )
		if ( updateInterval < 0 ) or ( updateInterval > 0.5 ) or not secondParm then
			printPC( "/tip u ".. TipDB.updateInterval.. pc_colour_Highlight.. " [ 0 <= ".. L["Update Interval"].. " <= 0.5]" )
		else
			TipDB.updateInterval = updateInterval
			printPC( L["Update Interval"].. " = ".. pc_colour_Highlight.. TipDB.updateInterval )
		end

	elseif ( options1 == "d" ) then
		if secondParm then
			printPC( "/tip d (".. ( ( TipDB.debug == "y" ) and L["On"] or L["Off"] ).. ")" )
		else
			TipDB.debug = ( TipDB.debug == "y" ) and "n" or "y"
			if ( TipDB.debug == "y" ) then
				printPC( L["Debug"].. ": ".. pc_colour_Highlight.. L["On"] )
			else
				printPC( L["Debug"].. ": ".. pc_colour_Highlight.. L["Off"] )
			end
		end

	elseif ( options1 == "v" ) then
		local version = GetAddOnMetadata( addonName, "Version" )
		printPC( L["Version: "].. pc_colour_Highlight.. version )

	-- Easter Eggs. Enjoy!

	elseif ( options1 == "z" ) then
		if ( soundkitHandle ) then
			StopSound( soundkitHandle, 5000 )
			soundkitHandle = nil
		else
			local soundSelection = random( 2 )
			if ( soundSelection == 1 ) then
				_, soundkitHandle = PlaySound( 78803 )		-- Tristram
			else
				local faction = UnitFactionGroup( "player" )
				if ( faction == "Alliance" ) then
					_, soundkitHandle = PlaySound( 79294 )	-- Winter Veil
				elseif ( faction == "Horde" ) then
					_, soundkitHandle = PlaySound( 79122 )
				else
					_, soundkitHandle = PlaySound( 79123 )
				end
			end
		end
	end
end

--print("n1="..(tostring(n1) or "nil").. " n2="..(tostring(n2) or "nil").. " n3="..(tostring(n3) or "nil").. " n4="..(tostring(n4) or "nil").. 
--		" n5="..(tostring(n5) or "nil").. " n6="..(tostring(n6) or "nil").. " n7="..(tostring(n7) or "nil").. 
--		" n8="..(tostring(n8) or "nil") ) 

--for k,v in pairs( owner.visualInfo ) do
--	print("k="..k.." v="..(tostring(v) or "nil"))
--end
--print(gsub(link, "\124", "\124\124"))
--|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*):?(%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?
